 /*
  * jiten.cpp - Dictionary Lookup Object
  * Copyright (c)2004 Pawel Szymczykowski <pawel@makenai.net>
  *
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
  *
  * It is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with Foobar; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */

#include "jiten.h"
#include "gpstdlib.h"
#include "gpstdio.h"
#include "gpmem.h"

#define JI_VER 15

Jiten::Jiten()
{
	this->reset();
}

Jiten::~Jiten()
{
	gm_free(this->index);
	gm_free(this->dictionary);
}

int Jiten::load(char *dict_filename, char *idx_filename)
{
	// Load entire dictionary and index into memory.
	// Check if index & dictionary match.
	// Returns non-zero on error.

	F_HANDLE fp;
	unsigned long m_size;

	GpFatInit();

	if (GpRelativePathSet("gp:\\GPMM\\GPNESE") != SM_OK)
	{
		return(1);
	}
	
	if (GpFileGetSize(dict_filename, &this->dictionary_length) != SM_OK)
	{
		return(2);
	}

	// Load Dictionary Data
	if (GpFileOpen(dict_filename, OPEN_R, &fp) != SM_OK)
	{
		//fprintf(stderr, "Unable to open dictionary data.\n");
		return(3);
	}

	this->dictionary = (unsigned char *) gm_malloc(this->dictionary_length + 2);
	if (this->dictionary == NULL)
	{
		//fprintf(stderr, "Could not allocate memory to store dictionary data.\n");
		GpFileClose(fp);
		return(4);	
	}
	this->dictionary[0] = '\n';
	if (GpFileRead(fp, this->dictionary + 1, this->dictionary_length, &m_size) != SM_OK)
	{
		//fprintf(stderr, "Error reading dictionary data.\n");
		GpFileClose(fp);
		return(5);	
	}
	this->dictionary[this->dictionary_length + 1] = '\n'; // Just in case.
	GpFileClose(fp);

	if (GpFileGetSize(idx_filename, &this->index_length) != SM_OK)
	{
		return(6);
	}

	// Load Dictionary Index
	if (GpFileOpen(idx_filename, OPEN_R, &fp) != SM_OK)
	{
		//fprintf(stderr, "Unable to open dictionary index.\n");
		return(7);
	}

	this->index_first_element = 1;
	this->index_last_element = this->index_length / sizeof(unsigned long) - 1;
	this->index = (unsigned long *) gm_malloc(this->index_length);
	if (this->index == NULL)
	{
		//fprintf(stderr, "Could not allocate memory to store dictionary index.\n");
		GpFileClose(fp);
		return(8);	
	}
	if (GpFileRead(fp, this->index, this->index_length, &m_size) != SM_OK)
	{
		//fprintf(stderr, "Error reading dictionary index.\n");
		GpFileClose(fp);
		return(9);	
	}
	GpFileClose(fp);

	// Check  if Index & Dictionary match
	if ((this->dictionary_length + JI_VER) != index[0])
	{
		//fprintf(stderr, "Index and dictionary data file do not match.\n");
		return(10);
	}

	return(0);
}

int Jiten::kstrncmp(const char *s1, const char *s2, int n)
{
	// effectively does a strnicmp on two "strings" 
	// except it will make katakana and hiragana match (EUC A4 & A5)
	unsigned c1,c2;
	int i,rc1,rc2;
	extern long it;

	for (i = 0; i<n ; i++)
	{
		c1 = s1[i];
		c2 = s2[i];
		if ((c1 == '\0')||(c2 == '\0')) return(0);
		if ((i % 2) == 0)
		{
			if (c1 == 0xA5)
			{
				c1 = 0xA4;
			}
			if (c2 == 0xA5)
			{
				c2 = 0xA4;
			}
		}
		if ((c1 >= 'A') && (c1 <= 'Z')) c1 |= 0x20; // fix ucase
		if ((c2 >= 'A') && (c2 <= 'Z')) c2 |= 0x20;
		if (c1 != c2 ) 
		{
			rc1 = c1;
			rc2 = c2;
			return(rc1-rc2);
		}
	}
	return(0);
}


int Jiten::search(char *term)
{
	// Search the dictionary via binary search, finding a match for term.
	// Compare backwards, looking for the first match.
	// Set index_current_element to the first match.
	// Returns 0 on match, 1 on failure.

	int lo = this->index_first_element;
	int hi = this->index_last_element;
	int md;
	int res;
	
	this->reset();
	while (1)
	{
		md = (lo+hi)/2;
		if ((res = this->kstrncmp(term, (const char *) this->dictionary + this->index[md], gm_lstrlen(term))) == 0)
		{
			this->search_term = term;
			this->index_current_element = md;
			this->index_current_element = this->first_match();	// rewind to the first match
			return 0;
		}

		if (res < 0) hi = md - 1;
		else lo = md + 1;
		if (lo > hi) return(1);
	}
	return(1);
}

int Jiten::num_matches()
{
	// return number of matches for current search term.
	// *** Problem.. will return incorrectly when the same term appears twice. See 'elephant'
	if (this->index_current_element == -1) return 0;
	return( (this->last_match() - this->first_match()) + 1 );
}

int Jiten::first_match()
{
	// returns the index of the first match for search_term

	int pos = this->index_current_element;
	if (pos == -1) return 0;


	while (1)
	{
		if(this->kstrncmp(this->search_term, (const char *) this->dictionary + this->index[pos], gm_lstrlen(this->search_term)) == 0)
		{
			pos--;
			if (pos == 0)
			{
				pos = 1;
				break;
			}
		} else {
			pos++;
			break;
		}
	}
	return(pos);
}

int Jiten::last_match()
{
	// returns the index of the last (final) match for search_term

	int pos = this->index_current_element;
	if (pos == -1) return 0;

	while (1)
	{
		if(this->kstrncmp(this->search_term, (const char *) this->dictionary + this->index[pos], gm_lstrlen(this->search_term)) == 0)
		{
			pos++;
			if (pos == this->index_last_element)
			{
				pos = this->index_last_element;
				break;
			}
		} else {
			pos--;
			break;
		}
	}
	return(pos);
}

void Jiten::reset()
{
	this->index_current_element = -1;
	this->search_term = NULL;
}

int Jiten::get_record(char * definition)
{
	// Get the record bounds and return a pointer to a new string containing it.
	// This could be buggy.. not sure. Was written quickly.. check up on it.
	unsigned long start = this->index[this->index_current_element];
	unsigned long end = this->index[this->index_current_element];
	int record_length;
	int i, j=0, fs=1;
	
	while (this->dictionary[start] != '\n') start--;
	start++;

	while (this->dictionary[end] != '\n') end++;
	end--;

	record_length = end - start + 1;
	
	for (i=0;i<record_length;i++)
	{
		if (this->dictionary[start + i] == '/')
		{
			if (fs) fs=0; 
			else {
				definition[j++] = ';';
				definition[j++] = ' ';
			}
		} else definition[j++] = this->dictionary[start + i];
	}
	definition[j - 2] = '\0';

	return 0;
}

int Jiten::next()
{
	// Increment the record pointer to the next hit.. return 1 if there are no more.

	if(this->kstrncmp(this->search_term, (const char *) this->dictionary + this->index[ this->index_current_element + 1 ], 
		gm_lstrlen(this->search_term)) == 0)
	{
		this->index_current_element++;
		return 0;
	}
	return 1;
}
